import pyglet
import math
import random

import image
import sound
import random
from constants import *

bullet_image = image.load('bullet')
bullet_image.anchor_x = bullet_image.width / 2
bullet_image.anchor_y = bullet_image.height / 2

life_image = image.load('life')

enemy_image = image.load('enemy')
enemy_hit_sound = sound.load('claping-ses1') #hit
hit_sound_semaphore = 5

# sounds taken from http://www.freesound.org/

sounds = [
           sound.load("synapse__trumpetloop01"),
           sound.load("synapse__trumpetloop02"),
           sound.load("simondsouza__innocent-falling-8beat-dmaj-110"),
           sound.load("simondsouza__uh-huh-that-s-funky-16beat-dmaj-110"),
           sound.load("boom-boom-8-beat-dm-90"),
           sound.load("boom-boom2-8-beat-dm-90"),
           sound.load("double-notes-8-beat-dm-90"),
           sound.load("a-e__inception-3")
         ]

sounds_semaphore = [1] * len(sounds)

# sounds = [sound_violin_6, sound_violin_14, sound_violin_11, sound_violin_12, sound_violin_9, sound_violin_15, sound_violin_8, sound_violin_7]

def get_distance(a, b):
    x_dist = b.x - a.x
    y_dist = b.y - a.y
    return math.hypot(x_dist, y_dist)

def get_closest_target(sprite):
    return min(enemy_manager.enemies, key=lambda enemy: get_distance(sprite, enemy))

map_right_turns = set()
map_down_turns = set()


class LifeManager(object):
    def __init__(self):
        self.lives = START_LIVES

    def decrement(self):
        self.lives -= 1

    def draw(self):
        for i in range(self.lives):
            life_image.blit(MAP_WIDTH - (i + 1) * (life_image.width + LIFE_GAP), MAP_HEIGHT - life_image.height)


life_manager = LifeManager()


class ScoreManager(pyglet.text.Label):
    def __init__(self):
        self.score = START_SCORE
        super(ScoreManager, self).__init__(str(self.score), FONTS, 30, color=(0, 0, 0, 255), y=2, x=640, anchor_x="right")

    def update_label(self):
        self.text = str(self.score)

    def increment(self):
        self.score += 1
        self.update_label()

    def purchase(self, cost):
        self.score -= cost
        self.update_label()


score_manager = ScoreManager()


class Bullet(pyglet.sprite.Sprite):
    def __init__(self, target, power, x, y, rotation, **kwargs):
        super(Bullet, self).__init__(bullet_image, **kwargs)

        self.target = target
        self.power = power
        self.x = x
        self.y = y
        self.rotate_to_target()

    def rotate_to_target(self):
        if not self.target:
            return

        x_dist = (self.target.x + self.target.width / 2) - self.x
        y_dist = (self.target.y + self.target.height / 2) - self.y

        self.rotation = math.degrees(math.atan2(x_dist, y_dist))

        self.xspeed = math.sin(math.radians(self.rotation))
        self.yspeed = math.cos(math.radians(self.rotation))

    def update(self):
        self.rotate_to_target()
        self.x += self.xspeed * BULLET_SPEED
        self.y += self.yspeed * BULLET_SPEED


class BulletManager(object):
    def __init__(self):
        self.batch = pyglet.graphics.Batch()
        self.bullets = []

    def add_bullet(self, target, power, x, y, rotation):
        bullet = Bullet(target, power, x, y, rotation, batch=self.batch)
        target.incoming_bullets.append(bullet)
        self.bullets.append(bullet)

    def remove_out_bullets(self):
        self.bullets = filter(lambda bullet: bullet.x > 0 and bullet.y > 0 and bullet.x < MAP_WIDTH and bullet.y < MAP_HEIGHT, self.bullets)

    def update(self):
        self.remove_out_bullets()
        for bullet in self.bullets:
            bullet.update()

        for enemy in enemy_manager.enemies:
            for bullet in self.bullets:
                if bullet.x >= enemy.x and bullet.y >= enemy.y and bullet.x <= enemy.x + enemy.width and bullet.y <= enemy.y + enemy.height:
                    enemy.hit(bullet.power)
                    self.bullets.remove(bullet)

                    # Remove references to the bullet
                    if bullet.target:
                        bullet.target.incoming_bullets.remove(bullet)

                    if enemy.dead:
                        try:
                            enemy_manager.enemies.remove(enemy)
                            enemy_manager.last_wave_enemies.append(EnemyStats(x = enemy.x,
                                                                              y = enemy.y,
                                                                              max_health = enemy.max_health,
                                                                              max_armour = enemy.max_armour,
                                                                              speed = enemy.orig_speed))
                            score_manager.increment()
                        except:
                            pass

    def draw(self):
        self.batch.draw()


bullet_manager = BulletManager()


class Tower(pyglet.sprite.Sprite):
    def __init__(self, *args, **kwargs):
        super(Tower, self).__init__(*args, **kwargs)

        self.x += (self.width / 2) - (self.x % TILE_SIZE)
        self.y += (self.height / 2) - (self.y % TILE_SIZE)

        self.last_shoot_delta = SHOOT_TIME

        self.power = {"health": 1,
                      "armour": 1,
                      "speed" : 1}

    def shoot(self, target):
        bullet_manager.add_bullet(target, self.power, self.x, self.y, self.rotation)
        self.last_shoot_delta = 0

    def update(self):
        self.last_shoot_delta += INTERVAL

        if enemy_manager.enemies:
            target = get_closest_target(self)
            x_dist = target.x + (target.width / 2) - self.x
            y_dist = target.y + (target.height / 2) - self.y

            if math.hypot(x_dist, y_dist) < TOWER_RANGE:
                self.rotation = math.degrees(math.atan2(x_dist, y_dist))

                if self.last_shoot_delta > SHOOT_TIME:
                    self.shoot(target)


class HealthTower(Tower):
    tower_image = image.load('health_tower', anchor_x=TILE_SIZE / 2, anchor_y=TILE_SIZE / 2)
    active_color = ACTIVE_HEALTH
    inactive_color = INACTIVE_HEALTH

    def __init__(self, *args, **kwargs):
        super(HealthTower, self).__init__(HealthTower.tower_image, *args, **kwargs)

        self.power['health'] = 2
        self.power['armour'] = 0.1
        self.power['speed'] = 0

        self.color = ACTIVE_HEALTH

class ArmorTower(Tower):
    tower_image = image.load('armor_tower', anchor_x=TILE_SIZE / 2, anchor_y=TILE_SIZE / 2)
    active_color = ACTIVE_ARMOUR
    inactive_color = INACTIVE_ARMOUR

    def __init__(self, *args, **kwargs):
        super(ArmorTower, self).__init__(ArmorTower.tower_image, *args, **kwargs)

        self.power['health'] = 0.1
        self.power['armour'] = 2
        self.power['speed'] = 0

        self.color = ACTIVE_ARMOUR


class SpeedTower(Tower):
    tower_image = image.load('speed_tower', anchor_x=TILE_SIZE / 2, anchor_y=TILE_SIZE / 2)
    active_color = ACTIVE_SPEED
    inactive_color = INACTIVE_SPEED

    def __init__(self, *args, **kwargs):
        super(SpeedTower, self).__init__(SpeedTower.tower_image, *args, **kwargs)

        self.power['health'] = 0.15
        self.power['armour'] = 0.15
        self.power['speed'] = 0.5

        self.color = ACTIVE_SPEED

class TowerManager(object):
    def __init__(self):
        self.batch = pyglet.graphics.Batch()

        self.towers = []

    def position_occupied(self, x, y):
        for tower in self.towers:
            if tower.x // TILE_SIZE == x and tower.y // TILE_SIZE == y:
                return True

        return False

    def add_tower(self, tower_class, x, y):
        if score_manager.score < TOWER_COST:
            return

        score_manager.purchase(TOWER_COST)
        tower = tower_class(batch=self.batch, x=x, y=y)
        self.towers.append(tower)

    def update(self):
        for tower in self.towers:
            tower.update()

    def draw(self):
        self.batch.draw()


tower_manager = TowerManager()

class EnemyStats():
    def __init__(self, x, y, max_health, max_armour, speed):
        if (x < 320):
            # First bit of the leg
            self.fitness = y + x
        else:
            # Second bit of the leg
            self.fitness = x - y + 1000
        self.max_health = max_health
        self.max_armour = max_armour
        self.speed = speed


class ThickLineGroup(pyglet.graphics.Group):
    def set_state(self):
        pyglet.gl.glLineWidth(3)


class Enemy(pyglet.sprite.Sprite):
    def __init__(self, max_health, max_armour, speed, **kwargs):
        super(Enemy, self).__init__(enemy_image, **kwargs)
        self.y = 0 - self.height

        max_stat = max(max_health, max_armour, speed)
        self.color = (255 * max_health / max_stat, 255 * speed / max_stat, 255 * max_armour / max_stat)

        group = ThickLineGroup()
        self.healthbar = self.batch.add(2, pyglet.gl.GL_LINES, group,
                                        ('v2f', (self.x, self.y+self.height+1, self.x + 20, self.y + self.height+1)),
                                        ('c3B', (255,195,194,255,0,0)))

        self.armourbar = self.batch.add(2, pyglet.gl.GL_LINES, group,
                                        ('v2f', (self.x, self.y+self.height+3, self.x + 20, self.y + self.height+3)),
                                        ('c3B', (0,0,255,0,0,255)))

        self.batch.draw()

        self.max_health = float(max_health)
        self.max_armour = float(max_armour)
        self.health = float(self.max_health)
        self.armour = float(self.max_armour)

        self.dead = False
        self.zone = -1

        self.incoming_bullets = []

        self.orig_speed = speed
        self.speed = speed

        self.direction = 'UP'

        self.distanceTravelled = 0

    def remove_indicator_bars(self):
        try:
            self.healthbar.delete()
            self.armourbar.delete()
        except AssertionError:
            pass

    def remove_bullets(self):
        # Remove references to the target
        for bullet in self.incoming_bullets:
            bullet.target = None

        self.incoming_bullets = []

    def hit(self, power):
        if self.dead:
            return

        global hit_sound_semaphore
        if (hit_sound_semaphore > 0):
            tempPlayer = enemy_hit_sound.play()
            hit_sound_semaphore -= random.randint(5,15)
        
            try:
                tempPlayer.volume = 0.1
            except:
                # No idea
                pass
        else:
            hit_sound_semaphore += 1 # todo: work out when the sound actually ended


        self.armour = max(self.armour - power["armour"], 0)
        healthHit = power["health"] - (0.9 * power["health"] * (self.armour / self.max_armour))
        self.health = max(self.health - healthHit, 0)

        if power['speed'] > 0:
            self.speed = self.orig_speed - float(power['speed'])
            pyglet.clock.unschedule(self.restore_speed)
            pyglet.clock.schedule_once(self.restore_speed, SLOW_SPEED_TIME)

        self.dead = self.health <= 0

        if self.dead:
            pyglet.clock.unschedule(self.restore_speed)
            self.remove_bullets()
            self.remove_indicator_bars()

    def restore_speed(self, dt):
        self.speed = self.orig_speed

    def update(self):
        if self.dead:
            return

        movementDelta = ENEMY_SPEED * self.speed

        if self.direction == 'UP':
            self.y += movementDelta
        elif self.direction == 'RIGHT':
            self.x += movementDelta
        elif self.direction == 'DOWN':
            self.y -= movementDelta

        self.distanceTravelled += movementDelta

        self.updateMusic()
        self.healthbar.vertices = [self.x, self.y+self.height+1, self.x+20*(self.health / self.max_health), self.y+self.height+1]
        self.armourbar.vertices = [self.x, self.y+self.height+3, self.x+20*(self.armour / self.max_armour), self.y+self.height+3]

        x_grid = self.x // TILE_SIZE
        y_grid = self.y // TILE_SIZE

        if y_grid >= 9: # speed hack
            if (x_grid, y_grid) in map_right_turns:
                self.direction = 'RIGHT'
            elif (x_grid, y_grid) in map_down_turns:
                self.direction = 'DOWN'

        if self.y <= -self.height:
            life_manager.decrement()
            self.dead = True
            self.remove_bullets()
            self.remove_indicator_bars()

    def updateMusic(self):
        # this controls how often sounds are played - in arch loop enemies can travel about 800 units
        diceval = random.randint(1,max(850 - int(self.distanceTravelled), 1))
        if (diceval == 1):

            # this distributes sounds over the screen
            newzone = int((self.distanceTravelled / 138))
            if (newzone < len(sounds) and newzone > self.zone):
                self.zone = newzone
                if (sounds_semaphore[newzone] > 0):
                    tempPlayer = sounds[newzone].play()
                    tempPlayer.volume = random.uniform(0.4,1)
                    sounds_semaphore[newzone] -= random.randint(1,5)
                else:
                    sounds_semaphore[newzone] += 1  # todo: work out when the sound actually ended


class EnemyManager(object):
    def __init__(self):
        self.batch = pyglet.graphics.Batch()

        self.enemies = []
        self.last_wave_enemies = []
        self.wave_time = WAVE_TIME
        self.spawn_wave(0)

    def spawn_wave(self, dt):
        self.wave_time = max(self.wave_time - WAVE_TIME_DEC, 1)
        pyglet.clock.schedule_once(self.spawn_wave, WAVE_TIME)
        # Setup the last_wave_enemies - limit this list to a sensible size (say, 10 or 20, try and hang onto the best)
        self.last_wave_enemies += map(lambda enemy: EnemyStats(x=enemy.x,
                                                               y=enemy.y,
                                                               max_health=enemy.max_health,
                                                               max_armour=enemy.max_armour,
                                                               speed=enemy.orig_speed),
                                      filter(lambda enemy: enemy.dead, self.enemies))

        if self.last_wave_enemies:
            best = max(self.last_wave_enemies, key=lambda stats: stats.fitness)
            print "Best: hp %f ap %f speed %f" % (best.max_health, best.max_armour, best.speed)

        if len(self.last_wave_enemies) > MAX_POP_SIZE:
            # Randomly remove some to get down to max pop size.
            random.shuffle(self.last_wave_enemies)
            self.last_wave_enemies = self.last_wave_enemies[0:MAX_POP_SIZE / 2]

        # Spawn each enemy separately
        for column in range(4):
            row_delays = range(0, int(WAVE_TIME * 1.5), 2)
            random.shuffle(row_delays)
            for delay in row_delays[:4]:
                pyglet.clock.schedule_once(self.spawn_enemy, delay, column)

    def spawn_enemy(self, dt, column):
        if len(self.last_wave_enemies) < TOURNAMENT_SIZE:
            self.enemies.append(Enemy(batch=self.batch,
                                      x=TILE_SIZE * (START_COLUMN + column),
                                      max_health=1,
                                      max_armour=1,
                                      speed=1))
        else:
            # Select a tournament of individuals from the last wave
            random.shuffle(self.last_wave_enemies)
            tournament = self.last_wave_enemies[0:TOURNAMENT_SIZE]

            # Find the best enemy in your tournament
            individual = max(tournament, key= lambda enemy: enemy.fitness)

            # Mutate that enemy by increasing it's max health or armour
            # Or flipping the stats
            randomNum = random.random()
            if randomNum < 0.3:
                self.enemies.append(Enemy(batch=self.batch,
                                          x=TILE_SIZE * (START_COLUMN + column),
                                          max_health=individual.max_health+2,
                                          max_armour=individual.max_armour,
                                          speed=individual.speed))
            elif randomNum < 0.6:
                self.enemies.append(Enemy(batch=self.batch,
                                          x=TILE_SIZE * (START_COLUMN + column),
                                          max_health=individual.max_health,
                                          max_armour=individual.max_armour+2,
                                          speed=individual.speed))
            elif randomNum < 0.9:
                self.enemies.append(Enemy(batch=self.batch,
                                          x=TILE_SIZE * (START_COLUMN + column),
                                          max_health=individual.max_health,
                                          max_armour=individual.max_armour,
                                          speed=individual.speed+ENEMY_SPEED_INC))
            else:
                # Uber mutation!
                self.enemies.append(Enemy(batch=self.batch,
                                          x=TILE_SIZE * (START_COLUMN + column),
                                          max_health=individual.max_armour+2,
                                          max_armour=individual.max_health+2,
                                          speed=individual.speed+ENEMY_SPEED_INC))

    def update(self):
        for enemy in self.enemies:
            enemy.update()

        # Remove references to dead enemies
        self.enemies = filter(lambda enemy: not enemy.dead, self.enemies)

    def draw(self):
        self.batch.draw()


enemy_manager = EnemyManager()
