import pyglet
from pyglet.gl import *

import entities
import gui
import image
import tile_map
import sound
from constants import *

window = pyglet.window.Window()

# Map
map1 = tile_map.Map('map1')
entities.map_right_turns = map1.get_right_turns() # not the best
entities.map_down_turns = map1.get_down_turns()

# Cursor
cursor = pyglet.sprite.Sprite(entities.SpeedTower.tower_image)
cursor.opacity /= 2

# Radius
radius_image = image.load('radius')
radius_image.anchor_x = radius_image.width / 2
radius_image.anchor_y = radius_image.height / 2
radius = pyglet.sprite.Sprite(radius_image)
radius.opacity /= 4

# GUI
gui_screen = gui.GUI()

# Enable alpha blending
glEnable(GL_BLEND)
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

# looping background music
sound_raggaman = sound.load("raggaman__3er-ding-0")

player = pyglet.media.Player()
player.eos_action = pyglet.media.Player.EOS_LOOP
player.queue(sound_raggaman)
player.volume = 0.2

def can_place_tower(x, y):
    grid_x = x // TILE_SIZE
    grid_y = y // TILE_SIZE
    return map1.is_wall(grid_x, grid_y) and not entities.tower_manager.position_occupied(grid_x, grid_y)

@window.event
def on_mouse_press(x, y, button, modifiers):
    gui_screen.on_mouse_press(x, y)
    cursor.image = gui_screen.selected.image
    if can_place_tower(x, y):
        entities.tower_manager.add_tower(gui_screen.selected.tower, x, y)

@window.event
def on_mouse_motion(x, y, dx, dy):
    cursor.x = x + (cursor.width / 2) - (x % TILE_SIZE)
    cursor.y = y + (cursor.height / 2) - (y % TILE_SIZE)
    radius.x = cursor.x
    radius.y = cursor.y

    cursor.visible = can_place_tower(x, y)
    radius.visible = cursor.visible

    gui_screen.on_mouse_motion(x, y, dx, dy)

# Fixed timestep loop
time_since_last_update = 0
def update(dt):
    global time_since_last_update
    
    # keep looping background music
    player.volume = 0.2
    player.play()    

    time_since_last_update += dt
    while time_since_last_update >= INTERVAL:
        cursor.rotation += 2

        entities.enemy_manager.update()
        entities.tower_manager.update()
        entities.bullet_manager.update()

        if entities.life_manager.lives < 0:
            # TODO: Add a 'game over' screen
            print "Game over man! Game over!"
            pyglet.app.exit()
            break

        time_since_last_update -= INTERVAL

pyglet.clock.schedule_interval(update, INTERVAL)

@window.event
def on_draw():
    window.clear()

    map1.draw()
    entities.enemy_manager.draw()
    entities.tower_manager.draw()
    entities.bullet_manager.draw()
    radius.draw()
    cursor.draw()

    gui_screen.draw()

    entities.score_manager.draw()
    entities.life_manager.draw()
