import pyglet

import image
import entities 

from constants import *

towers = {0: entities.SpeedTower,
          1: entities.ArmorTower,
          2: entities.HealthTower}

class GUI(object):
    def __init__(self):
        self.x = MAP_WIDTH
        self.y = 0
        self.width = TILE_SIZE * 2
        self.height = MAP_HEIGHT

        background_image = image.load('sidebar')
        self.background = pyglet.image.TileableTexture.create_for_image(background_image)

        self.button_batch = pyglet.graphics.Batch()
        self.buttons = [self.create_button(i) for i in range(3)]

        self.hovered = None
        self.selected = self.buttons[0]
        self.selected.color = self.selected.tower.active_color

    def create_button(self, i):
        button_image = towers[i].tower_image
        button_image.anchor_x = button_image.width / 2
        button_image.anchor_y = button_image.height / 2

        x_pos = self.x + (self.width / 2)
        y_pos = self.height - button_image.height - button_image.height * i * 2

        sprite = pyglet.sprite.Sprite(button_image, batch=self.button_batch, x=x_pos, y=y_pos)
        sprite.tower = towers[i]
        sprite.color = sprite.tower.inactive_color
        return sprite

    def over_button(self, x, y):
        for button in self.buttons:
            bounds = [
                button.x - (button.width / 2),
                button.x + (button.width / 2),
                button.y - (button.height / 2),
                button.y + (button.height / 2)
            ]

            if x > bounds[0] and x < bounds[1] and y > bounds[2] and y < bounds[3]:
                return button

    def on_mouse_motion(self, x, y, dx, dy):
        sprite = self.over_button(x, y)

        if self.hovered and self.hovered != self.selected:
            self.hovered.color = self.hovered.tower.inactive_color

        if sprite:
            self.hovered = sprite
            sprite.color = sprite.tower.active_color

    def on_mouse_press(self, x, y):
        sprite = self.over_button(x,y)

        if sprite:
            if self.selected:
                self.selected.color = self.selected.tower.inactive_color
            self.selected = sprite
            sprite.color = sprite.tower.active_color

    def draw(self):
        self.background.blit_tiled(self.x, self.y, 0, self.width, self.height)
        self.button_batch.draw()
