import pyglet
from collections import defaultdict

import image
from constants import *

WALL =       (chr(255), chr(0),   chr(0))
PATH =       (chr(0),   chr(0),   chr(0))
RIGHT_TURN = (chr(0),   chr(255), chr(0))
DOWN_TURN =  (chr(0),   chr(255), chr(255))
BLOCKER =    (chr(255), chr(255), chr(255))

wall_image = image.load('tile1')
path_image = image.load('tile2')
blocker_image = image.load('blocker')


class Map(object):
    def __init__(self, name):
        map_image = image.load('map1')
        self.data = map_image.get_data('RGB', map_image.width * 3)
        self.batch = pyglet.graphics.Batch()
        self.sprites = defaultdict(lambda: {})

        self.right_turns = set()
        self.down_turns = set()

        self.load(map_image)

    def load(self, map_image):
        for y in range(map_image.height):
            x_lower = y * map_image.width * 3
            x_upper = (y + 1) * map_image.width * 3
            row = self.data[x_lower:x_upper]
            for x, pixel in enumerate(zip(row[0::3], row[1::3], row[2::3])):
                if pixel == WALL:
                    tile_image = wall_image
                elif pixel == PATH:
                    tile_image = path_image
                elif pixel == BLOCKER:
                    tile_image = blocker_image
                elif pixel == RIGHT_TURN:
                    tile_image = path_image
                    self.right_turns.add((x, y))
                elif pixel == DOWN_TURN:
                    tile_image = path_image
                    self.down_turns.add((x, y))
                else:
                    raise Exception("Pixel colour (%s) doesn't map to tile" % str(pixel))
                x_pos = x * TILE_SIZE
                y_pos = y * TILE_SIZE
                sprite = pyglet.sprite.Sprite(tile_image, batch=self.batch, x=x_pos, y=y_pos)
                self.sprites[x][y] = sprite

    def is_tile_of_texture(self, x, y, texture):
        try:
            if self.sprites[x][y].image is texture:
                return True
        except:
            pass

        return False

    def get_right_turns(self):
        return self.right_turns

    def get_down_turns(self):
        return self.down_turns

    def is_wall(self, x, y):
        return self.is_tile_of_texture(x, y, wall_image.texture)

    def draw(self):
        self.batch.draw()
